unit Unit1;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, IB_Components, StdCtrls, IBOServices, ComCtrls, IB_Process,
  IB_Script;

type
  TForm1 = class(TForm)
    CurrentDB: TIB_Connection;
    Control: TIB_Cursor;
    Label1: TLabel;
    edDBPath: TEdit;
    Backup: TIBOBackupService;
    Restore: TIBORestoreService;
    btConnect: TButton;
    ScriptV2: TIB_Script;
    trSnapshot: TIB_Transaction;
    ScriptV3: TIB_Script;
    MemoLog: TMemo;
    procedure CurrentDBAfterConnect(Sender: TIB_Connection);
    procedure FormCreate(Sender: TObject);
    procedure btConnectClick(Sender: TObject);
  private
    { Private declarations }
    procedure InitUpdateProccess;
    procedure EndUpdateProccess;
    procedure RestoreOriginalDB;
    procedure Log(s: string);
    procedure UpdateToVersion2;
    procedure UpdateToVersion3;
    procedure Reconnect;
  public
    { Public declarations }
  end;

var
  Form1: TForm1;

implementation

{$R *.dfm}

uses DateUtils;

const CurrentDBVersion = 3;

var
  BackupFileName: string;
  DBUpdating: Boolean;

procedure TForm1.CurrentDBAfterConnect(Sender: TIB_Connection);
var DBVersion: integer;
begin
  // If we are in the middle of a update proccess, just exit.
  // Sometimes, reconnect are necessary.
  if DBUpdating then exit;

  Control.First; // Open  cursor

  if Control.FieldByName('InMaintenance').asString <> 'N' then
  // DB is currently being updated... quit application.
  begin
    Control.Close;
    CurrentDB.CloseTransactions;
    CurrentDB.Disconnect;
    MessageDlg('Database is currently being updated. Wait a few minutes and try again...', mtWarning, [mbOk], 0);
    Halt;
  end;

  DBVersion := Control.FieldByName('dbversion').asInteger;

  // App is too old for database version
  if DBVersion > CurrentDBVersion then
  begin
    Control.Close;
    CurrentDB.CloseTransactions;
    CurrentDB.Disconnect;
    MessageDlg('Your database version is higher than the application version. Please update your application.', mtWarning, [mbOk], 0);
    Halt;
  end
  else
    // DB structure needs to be updated
    if DBVersion < CurrentDBVersion then
    begin
      InitUpdateProccess;
      try
        if DBVersion < 2 then UpdateToVersion2; // Check if DB is version 1
        if DBVersion < 3 then UpdateToVersion3; // Check if DB is version 2
        // when new updates are available, insert a new line here...
        EndUpdateProccess;
      except
        on msg: exception do
        begin
          MessageDlg('There was an error trying to update the database structure: ' + msg.Message + #13 + 'The original database will be now restored.', mtWarning, [mbOk], 0);
          RestoreOriginalDB;
          Application.terminate;
        end;
      end;
    end
    else
      Log('Database is up to date');
end;

procedure TForm1.FormCreate(Sender: TObject);
begin
  edDBPath.Text := ExtractFilePath(ParamStr(0)) + 'database.fdb';
end;

procedure TForm1.InitUpdateProccess;
var x: string;
begin
  // Verify if there is only our connection active
  if (CurrentDB.Users.count <> 1) then
  begin
    CurrentDB.Disconnect;
    MessageDlg('Database needs "exclusive" connection to be updated! Aborting...', mtWarning, [mbOk], 0);
    Application.Terminate;
    Abort;
  end;

  // Set the DB Maintenance flag
  with Control do
  begin
    First;
    if FieldByName('InMaintenance').asString <> 'N' then
    begin
      Close;
      CurrentDB.CloseTransactions;
      CurrentDB.Disconnect;
      MessageDlg('Database is currently being updated. Wait a few minutes and try again...', mtWarning, [mbOk], 0);
      Halt;
    end;
    try
      // Check again if another terminal is already DBUpdating the DB
      Edit;
      FieldByName('InMaintenance').asString := 'Y';
      Post;
    finally
      close;
    end;
  end;

  // Do backup copy of database to be able to recover it if any problem happens
  with Backup do
  begin
    Params.Clear;
    Params.Add('user_name=' + CurrentDB.Username);
    Params.Add('password=' + CurrentDB.Password);
    ServerName := CurrentDB.Server;
    DatabaseName := CurrentDB.Path;
    Backup.Protocol := CurrentDB.Protocol;
    BackupFileName := ExtractFilePath(CurrentDB.Path) + 'TmpCopy_' + IntToStr(MilliSecondOfTheDay(Now));
    BackupFile.Text := BackupFileName + '.gbk';
    Active := True;
    try
      Log('Doing backup...');
      try
        ServiceStart;
        while not Eof do
        begin
          x := GetNextLine;
        end;
        while IsServiceRunning do
        begin
          Application.ProcessMessages;
          Screen.Cursor := crHourGlass;
        end;
      except
        ShowMessage('Problem during backup creation!');
        raise;
      end;
    finally
      active := false;
    end;
  end;

  DBUpdating := true;
end;

procedure TForm1.btConnectClick(Sender: TObject);
begin
  CurrentDB.Path:=edDBPath.Text;
  CurrentDB.Connect;
end;

procedure TForm1.EndUpdateProccess;
begin
  CurrentDB.CloseTransactions;
  with Control do
  begin
    First;
    try
      Edit;
      FieldByName('InMaintenance').asString := 'N';
      Post;
    finally
      close;
    end;
  end;
  DBUpdating := false;
  Log('Database structure is now up to date.');
end;

procedure TForm1.RestoreOriginalDB;
var x: string;
begin
  // Restore new DB
  CurrentDB.CloseTransactions;
  CurrentDB.Disconnect;
  Application.ProcessMessages;
  with Restore do
  begin
    PageSize := 4096;
    LoginPrompt := False;
    Params.Assign(Backup.Params);
    ServerName := Backup.ServerName;
    DatabaseName.Text := CurrentDB.path;
    Restore.Protocol := Backup.Protocol;
    BackupFile.Text := BackupFileName + '.gbk';
    Active := True;
    try
      Log('Restoring original DB...');
      try
        ServiceStart;
        while not Eof do
        begin
          x := GetNextLine;
        end;
        while IsServiceRunning do
        begin
          Application.ProcessMessages;
          Screen.Cursor := crHourGlass;
        end;
      except
        ShowMessage('It was not possible to restore the original database file. Its backup is saved in file ' + BackupFileName);
        raise;
      end;
    finally
      Active := False;
    end;
    Control.ExecuteDML('update control set InMaintenance = '#39'N'#39,nil);
  end;
end;

procedure TForm1.UpdateToVersion2;
begin
  Log('Upgrading to version 2');
  ScriptV2.execute;
  Reconnect;
end;

procedure TForm1.UpdateToVersion3;
begin
  Log('Upgrading to version 3');
  ScriptV3.execute;
  Reconnect;
end;

procedure TForm1.Log(s: string);
begin
  MemoLog.Lines.add('[' + DateTimeToStr(Now) + '] ' + s);
end;

procedure TForm1.Reconnect;
begin
  with CurrentDB do
  begin
    CloseTransactions;
    Disconnect;
    Connect;
  end;
end;

end.

